use std::{collections::HashMap, net::SocketAddr};

use anyhow::anyhow;
use async_tungstenite::tokio::connect_async;
use async_tungstenite::tungstenite::Message;
use futures_util::{SinkExt, StreamExt};
use lazy_static::lazy_static;
use log::{error, info, trace};
use serde::{Deserialize, Serialize};
use tokio::{
    io::{AsyncReadExt, AsyncWriteExt},
    net::{
        tcp::{OwnedReadHalf, OwnedWriteHalf},
        TcpStream,
    },
    sync::mpsc::{UnboundedReceiver, UnboundedSender},
    sync::RwLock,
    task,
};

enum ChannelHandlerMessage {
    Data(Vec<u8>),
    Shutdown,
}

lazy_static! {
    static ref CHANNEL_MAP: RwLock<HashMap<u8, UnboundedSender<ChannelHandlerMessage>>> =
        RwLock::new(HashMap::new());
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    let _ = env_logger::try_init();
    let args = Args::parse();
    let server = args.server.unwrap_or("wss://e4mc.skyevg.systems".to_string());
    let (ws_conn, _) = connect_async(server).await?;
    let (mut send, mut recv) = ws_conn.split();
    let (sender, mut receiver) = tokio::sync::mpsc::unbounded_channel();
    let sender = &*Box::leak(Box::new(sender)); // the task::spawn below requires 'static and this is main
    tokio::try_join!(
        async move {
            while let Some(Ok(message)) = recv.next().await {
                match message {
                    Message::Text(message) => {
                        info!("{}", message);
                        let message: ClientboundControlMessage = serde_json::from_str(&message)?;
                        match message {
                            ClientboundControlMessage::DomainAssigned(domain) => {
                                println!("Domain assigned: {}", domain);
                            }
                            ClientboundControlMessage::ChannelOpen(id, _) => {
                                match TcpStream::connect(("127.0.0.1", args.port)).await {
                                    Ok(conn) => {
                                        let (send, recv) = tokio::sync::mpsc::unbounded_channel();
                                        CHANNEL_MAP.write().await.insert(id, send);
                                        task::spawn(async move {
                                            if let Err(e) =
                                                handle_channel(id, sender.clone(), conn, recv).await
                                            {
                                                error!("Error handling channel: {}", e);
                                            }
                                            CHANNEL_MAP.write().await.remove(&id);
                                            info!("Sending close request for channel: {}", id);
                                            sender
                                                .send(Message::Text(
                                                    serde_json::to_string(
                                                        &ServerboundControlMessage::ChannelClosed(
                                                            id,
                                                        ),
                                                    )
                                                    .unwrap(),
                                                ))
                                                .unwrap();
                                        });
                                    }
                                    Err(e) => {
                                        error!("Error creating channel: {}", e);
                                        sender
                                            .send(Message::Text(
                                                serde_json::to_string(
                                                    &ServerboundControlMessage::ChannelClosed(
                                                        id,
                                                    ),
                                                )
                                                .unwrap(),
                                            ))
                                            .unwrap();
                                    }
                                };
                            }
                            ClientboundControlMessage::ChannelClosed(id) => {
                                let mut channel_map = CHANNEL_MAP.write().await;
                                if let Some(send) = channel_map.remove(&id) {
                                    if let Err(e) = send.send(ChannelHandlerMessage::Shutdown) {
                                        error!("Error closing channel: {}", e);
                                    }
                                }
                            }
                        }
                    }
                    Message::Binary(buf) => {
                        trace!("recv: {:?}", buf);
                        if let Some(send) = CHANNEL_MAP.write().await.get_mut(&buf[0]) {
                            if let Err(e) = send.send(ChannelHandlerMessage::Data(buf[1..].to_vec())) {
                                error!("Error transferring data: {}", e);
                            }
                        }
                    }
                    _ => {}
                }
            }
            Ok(()) as anyhow::Result<()>
        },
        async move {
            while let Some(message) = receiver.recv().await {
                send.send(message.clone()).await?;
            }
            Ok(()) as anyhow::Result<()>
        }
    )?;
    Ok(())
}

async fn handle_channel(
    id: u8,
    sender: UnboundedSender<Message>,
    mut conn: TcpStream,
    mut recv: UnboundedReceiver<ChannelHandlerMessage>,
) -> anyhow::Result<()> {
    let (mut read, mut write) = conn.split();
    tokio::select!(_ = async move {
        let mut buf = [0u8; 1024];
        loop {
            let ready = read.ready(tokio::io::Interest::READABLE).await?;
            if ready.is_read_closed() {
                return Ok(())
            }
            let len = read.read(&mut buf).await?;
            if len == 0 {
                continue;
            }
            let mut packet = Vec::from(&buf[..len]);
            packet.insert(0, id);
            trace!("send: {:?}", packet);
            sender.send(Message::Binary(packet))?;
        }
        Ok(()) as anyhow::Result<()>
    } => {},
    _ = async move {
        while let Some(message) = recv.recv().await {
            match message {
                ChannelHandlerMessage::Data(buf) => write.write_all(&buf).await?,
                ChannelHandlerMessage::Shutdown => return Ok(()),
            }
        }
        Ok(()) as anyhow::Result<()>
    } => {});
    conn.shutdown().await?;
    Ok(())
}

use clap::Parser;

/// Simple program to greet a person
#[derive(Parser, Debug)]
struct Args {
    /// Port of Minecraft server
    #[arg(short, long, default_value_t = 25565)]
    port: u16,

    /// Address of e4mc-server instance
    #[arg(short, long)]
    server: Option<String>,
}

#[derive(Deserialize)]
enum ClientboundControlMessage {
    DomainAssigned(String),
    ChannelOpen(u8, SocketAddr),
    ChannelClosed(u8),
}

#[derive(Serialize)]
enum ServerboundControlMessage {
    ChannelClosed(u8),
}
