use std::collections::{HashMap};
use std::env;
use std::net::SocketAddr;

use anyhow::{anyhow, Result};
use futures::{SinkExt, StreamExt};
use tokio::io::{AsyncReadExt, AsyncWriteExt};
use tokio::sync::mpsc::{UnboundedReceiver, UnboundedSender};
use tokio::net::{TcpListener, TcpStream};
use tokio::sync::RwLock;
use tokio::task;
use async_tungstenite::tungstenite::Message;
use lazy_static::lazy_static;
use log::{error, info};
use nanoid::nanoid;
use netty::ReadExtNetty;
use serde::{Deserialize, Serialize};

use crate::netty::{NettyReadError, WriteExtNetty};

mod netty;

#[derive(Debug, Clone)]
struct Handshake {
    protocol_version: i32,
    server_address: String,
    server_port: u16,
    next_state: HandshakeType,
}

#[derive(Debug, Clone)]
#[repr(i32)]
enum HandshakeType {
    Status = 1,
    Login = 2,
}

impl Handshake {
    async fn new(mut packet: &[u8]) -> Result<Self> {
        let packet_type = packet.read_varint().await?;
        if packet_type != 0 {
            Err(anyhow!("Not a Handshake packet"))
        } else {
            let protocol_version = packet.read_varint().await?;
            let server_address = packet.read_string().await?;
            let server_port = packet.read_unsigned_short().await?;
            let next_state = match packet.read_varint().await? {
                1 => HandshakeType::Status,
                2 => HandshakeType::Login,
                _ => return Err(anyhow!("Invalid next state")),
            };
            Ok(Self {
                protocol_version,
                server_address,
                server_port,
                next_state,
            })
        }
    }
}

lazy_static! {
    static ref EXPOSER_MAP: RwLock<HashMap<String, UnboundedSender<(Handshake, TcpStream, SocketAddr)>>> =
        RwLock::new(HashMap::new());
    static ref BASE_DOMAIN: String = env::var("BASE_DOMAIN").expect("BASE_DOMAIN missing");
}

#[tokio::main]
async fn main() -> Result<()> {
    let _ = env_logger::try_init();
    let ws_bind_addr = env::var("WS_BIND_ADDR").unwrap_or_else(|_| "127.0.0.1:80".to_string());
    let mc_bind_addr = env::var("MC_BIND_ADDR").unwrap_or_else(|_| "0.0.0.0:25565".to_string());

    futures::try_join!(
        async {
            let listener = TcpListener::bind(&ws_bind_addr).await?;
            info!("WebSocket Listening on: {}", ws_bind_addr);
            while let Ok((stream, _)) = listener.accept().await {
                task::spawn(async {
                    if let Err(e) = accept_ws_connection(stream).await {
                        error!("Error handling WebSocket connection: {}", e);
                    }
                });
            }
            Ok(()) as Result<()>
        },
        async {
            let listener = TcpListener::bind(&mc_bind_addr).await?;
            info!("Minecraft Listening on: {}", mc_bind_addr);
            while let Ok((stream, _)) = listener.accept().await {
                task::spawn(async {
                    if let Err(e) = accept_mc_connection(stream).await {
                        error!("Error handling Minecraft connection: {}", e);
                    }
                });
            }
            Ok(()) as Result<()>
        }
    )?;

    Ok(())
}

#[derive(Serialize)]
enum ClientboundControlMessage {
    DomainAssigned(String),
    ChannelOpen(u8, SocketAddr),
    ChannelClosed(u8),
}

#[derive(Deserialize)]
enum ServerboundControlMessage {
    ChannelClosed(u8),
}

struct ExposerMapHandle(String, UnboundedReceiver<(Handshake, TcpStream, SocketAddr)>);

impl ExposerMapHandle {
    async fn new(domain: String) -> Self {
        let (sender, receiver) = tokio::sync::mpsc::unbounded_channel();
        EXPOSER_MAP.write().await.insert(domain.clone(), sender);
        Self(domain, receiver)
    }

    async fn recv(&mut self) -> Option<(Handshake, TcpStream, SocketAddr)> {
        self.1.recv().await
    }
}

impl Drop for ExposerMapHandle {
    fn drop(&mut self) {
        let domain = self.0.clone();
        task::spawn(async move {
            EXPOSER_MAP.write().await.remove(&domain);
        });
    }
}

struct ChannelHandle(
    u8,
    UnboundedReceiver<MinecraftConnectionMessage>,
    UnboundedSender<WebsocketConnectionMessage>,
);

impl ChannelHandle {
    async fn new(
        id: u8,
        addr: SocketAddr,
        send: UnboundedSender<WebsocketConnectionMessage>,
    ) -> Result<Self> {
        let (sender, receiver) = tokio::sync::mpsc::unbounded_channel();
        send.send(WebsocketConnectionMessage::Open(id, addr, sender))
            .map_err(|e| anyhow::anyhow!("{}", e))?;
        Ok(Self(id, receiver, send))
    }

    async fn recv(&mut self) -> Option<MinecraftConnectionMessage> {
        self.1.recv().await
    }
}

impl Drop for ChannelHandle {
    fn drop(&mut self) {
        self.2.send(WebsocketConnectionMessage::Close(self.0)).map_err(|e| anyhow::anyhow!("{}", e)).unwrap();
    }
}

enum MinecraftConnectionMessage {
    Data(Vec<u8>),
    Close,
}

enum WebsocketConnectionMessage {
    Data(Vec<u8>),
    Close(u8),
    Open(u8, SocketAddr, UnboundedSender<MinecraftConnectionMessage>),
}

async fn accept_ws_connection(stream: TcpStream) -> Result<()> {
    let addr = stream.peer_addr()?;
    info!("WebSocker Peer address: {}", addr);

    let mut ws_stream = async_tungstenite::tokio::accept_async(stream).await?;

    info!("New WebSocket connection: {}", addr);

    let mut domain = get_random_domain();

    let exposer_map = EXPOSER_MAP.read().await;
    while exposer_map.contains_key(&domain) {
        domain = get_random_domain();
    }
    let domain = domain;
    info!("Connection {} was assigned domain {}", addr, domain);

    ws_stream
        .send(Message::Text(serde_json::to_string(
            &ClientboundControlMessage::DomainAssigned(domain.clone()),
        )?))
        .await?;

    drop(exposer_map);

    let mut exposer_handle = ExposerMapHandle::new(domain.clone()).await;

    let channel_map = &RwLock::new(HashMap::new());

    let (sender, mut receiver) = tokio::sync::mpsc::unbounded_channel();
    let (mut send, mut recv) = ws_stream.split();

    futures::try_join!(
        async move {
            while let Some(message) = receiver.recv().await {
                match message {
                    WebsocketConnectionMessage::Data(buf) => {
                        send.send(Message::Binary(buf)).await?
                    }
                    WebsocketConnectionMessage::Close(id) => {
                        send.send(Message::Text(serde_json::to_string(
                            &ClientboundControlMessage::ChannelClosed(id),
                        )?))
                        .await?;
                        channel_map.write().await.remove(&id);
                    }
                    WebsocketConnectionMessage::Open(id, addr, channel) => {
                        send.send(Message::Text(serde_json::to_string(
                            &ClientboundControlMessage::ChannelOpen(id, addr),
                        )?))
                        .await?;
                        channel_map.write().await.insert(id, channel);
                    }
                }
            }
            Ok(()) as Result<()>
        },
        async move {
            while let Some(Ok(message)) = recv.next().await {
                match message {
                    Message::Text(message) => {
                        let message: ServerboundControlMessage = serde_json::from_str(&message)?;
                        match message {
                            ServerboundControlMessage::ChannelClosed(id) => {
                                if let Some(channel) = channel_map.write().await.remove(&id) {
                                    info!("Closing channel id: {}", id);
                                    channel.send(MinecraftConnectionMessage::Close).map_err(|e| anyhow::anyhow!("{}", e))?;
                                }
                            }
                        }
                    }
                    Message::Binary(buf) => {
                        if let Some(channel) = channel_map.read().await.get(&buf[0]) {
                            channel
                                .send(MinecraftConnectionMessage::Data(buf[1..].to_vec())).map_err(|e| anyhow::anyhow!("{}", e))?;
                        }
                    }
                    _ => {}
                }
            }
            Ok(()) as Result<()>
        },
        async move {
            while let Some((handshake, mc_stream, addr)) = exposer_handle.recv().await {
                if let Some(channel_id) =
                    get_available_channel(channel_map.read().await.keys().copied().collect())
                {
                    task::spawn(handle_sent_connection(
                        handshake,
                        mc_stream,
                        addr,
                        sender.clone(),
                        channel_id,
                    ));
                }
            }
            Ok(()) as Result<()>
        }
    )?;
    Ok(())
}

async fn handle_sent_connection(
    handshake: Handshake,
    mut mc_stream: TcpStream,
    addr: SocketAddr,
    send: UnboundedSender<WebsocketConnectionMessage>,
    channel_id: u8,
) -> Result<()> {
    let mut handle = ChannelHandle::new(channel_id, addr, send.clone()).await?;
    let mut buf = vec![];
    buf.write_varint(0).await?;
    buf.write_varint(handshake.protocol_version).await?;
    buf.write_string(&handshake.server_address).await?;
    buf.write_all(&handshake.server_port.to_be_bytes()).await?;
    buf.write_varint(handshake.next_state as i32).await?;
    let mut len_buf = vec![channel_id];
    len_buf.write_varint(buf.len() as i32).await?;
    len_buf.append(&mut buf);
    send.send(WebsocketConnectionMessage::Data(len_buf)).map_err(|e| anyhow::anyhow!("{}", e))?;
    let (mut read, mut write) = mc_stream.split();
    tokio::select!(_ =
        #[allow(unreachable_code)]
        async move {
            let mut buf = [0u8; 1024];
            loop {
                read.readable().await?;
                let len = read.read(&mut buf).await?;
                if len == 0 {
                    continue;
                }
                let mut packet = Vec::from(&buf[..len]);
                packet.insert(0, channel_id);
                send.send(WebsocketConnectionMessage::Data(packet)).map_err(|e| anyhow::anyhow!("{}", e))?;
            }
            Ok(()) as anyhow::Result<()>
        } => {},
        _ = async move {
            while let Some(message) = handle.recv().await {
                match message {
                    MinecraftConnectionMessage::Data(buf) => {
                        write.write_all(&buf).await?;
                    }
                    MinecraftConnectionMessage::Close => {
                        return Ok(())
                    }
                }
            }
            Ok(()) as anyhow::Result<()>
        } => {}
    );
    mc_stream.shutdown().await?;
    Ok(())
}

async fn accept_mc_connection(mut stream: TcpStream) -> Result<()> {
    let addr = stream.peer_addr()?;
    info!("New Minecraft connection: {}", addr);

    let packet = stream.read_packet().await;

    if let Err(NettyReadError::LegacyServerListPing) = packet {
        stream
            .write_all(include_bytes!("./legacy_serverlistping_response.bin"))
            .await?;
        return Ok(());
    }

    let packet = packet?;

    let handshake = Handshake::new(&packet).await?;
    if let Some(sender) = EXPOSER_MAP.read().await.get(&handshake.server_address) {
        sender.send((handshake, stream, addr)).map_err(|e| anyhow::anyhow!("{}", e))?;
    } else {
        match handshake.next_state {
            HandshakeType::Status => {
                let mut buf = vec![];
                buf.write_varint(0).await?;
                buf.write_string(include_str!("./serverlistping_response.json"))
                    .await?;
                stream.write_varint(buf.len() as i32).await?;
                stream.write_all(&buf).await?;
            }
            HandshakeType::Login => {
                let _ = stream.read_packet().await?;
                let mut buf = vec![];
                buf.write_varint(0).await?;
                buf.write_string(include_str!("./disconnect_response.json"))
                    .await?;
                stream.write_varint(buf.len() as i32).await?;
                stream.write_all(&buf).await?;
            }
        }
    }
    Ok(())
}

const ID_ALPHABET: [char; 36] = [
    '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i',
    'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z',
];
const ID_LENGTH: usize = 8;

fn get_random_domain() -> String {
    format!(
        "{}.{}",
        nanoid!(ID_LENGTH, &ID_ALPHABET),
        BASE_DOMAIN.as_str()
    )
}

fn get_available_channel(used: Vec<u8>) -> Option<u8> {
    (0u8..=255).find(|&i| !used.contains(&i))
}
