use tokio::io::{AsyncReadExt, AsyncWriteExt};

use async_trait::async_trait;
use log::{error};
use thiserror::Error;

#[derive(Error, Debug)]
pub enum NettyReadError {
    #[error("{0}")]
    IoError(std::io::Error),
    #[error("Was not a netty packet, but a Legacy ServerListPing")]
    LegacyServerListPing,
}

impl From<std::io::Error> for NettyReadError {
    fn from(value: std::io::Error) -> Self {
        Self::IoError(value)
    }
}

impl From<std::io::ErrorKind> for NettyReadError {
    fn from(value: std::io::ErrorKind) -> Self {
        Self::IoError(value.into())
    }
}

#[async_trait]
pub trait ReadExtNetty: AsyncReadExt + Unpin {
    async fn read_bool(&mut self) -> Result<bool, NettyReadError> {
        let mut buf = [0u8];
        self.read_exact(&mut buf).await?;
        match buf[0] {
            0 => Ok(false),
            1 => Ok(true),
            _ => Err(std::io::ErrorKind::InvalidData.into()),
        }
    }

    async fn read_byte(&mut self) -> Result<i8, NettyReadError> {
        let mut buf = [0u8];
        self.read_exact(&mut buf).await?;
        Ok(i8::from_be_bytes(buf))
    }

    async fn read_unsigned_byte(&mut self) -> Result<u8, NettyReadError> {
        let mut buf = [0u8];
        self.read_exact(&mut buf).await?;
        Ok(buf[0])
    }

    async fn read_short(&mut self) -> Result<i16, NettyReadError> {
        let mut buf = [0u8; 2];
        self.read_exact(&mut buf).await?;
        Ok(i16::from_be_bytes(buf))
    }

    async fn read_unsigned_short(&mut self) -> Result<u16, NettyReadError> {
        let mut buf = [0u8; 2];
        self.read_exact(&mut buf).await?;
        Ok(u16::from_be_bytes(buf))
    }

    async fn read_int(&mut self) -> Result<i32, NettyReadError> {
        let mut buf = [0u8; 4];
        self.read_exact(&mut buf).await?;
        Ok(i32::from_be_bytes(buf))
    }

    async fn read_long(&mut self) -> Result<i64, NettyReadError> {
        let mut buf = [0u8; 8];
        self.read_exact(&mut buf).await?;
        Ok(i64::from_be_bytes(buf))
    }

    async fn read_float(&mut self) -> Result<f32, NettyReadError> {
        let mut buf = [0u8; 4];
        self.read_exact(&mut buf).await?;
        Ok(f32::from_be_bytes(buf))
    }

    async fn read_double(&mut self) -> Result<f64, NettyReadError> {
        let mut buf = [0u8; 8];
        self.read_exact(&mut buf).await?;
        Ok(f64::from_be_bytes(buf))
    }

    async fn read_string(&mut self) -> Result<String, NettyReadError> {
        let len = self.read_varint().await?;
        let mut buf = vec![0u8; len as usize];
        self.read_exact(&mut buf).await?;
        String::from_utf8(buf).map_err(|_| std::io::ErrorKind::InvalidData.into())
    }

    
    async fn read_varint(&mut self) -> Result<i32, NettyReadError> {
        let mut res = 0i32;
        for i in 0..5 {
            let part = self.read_unsigned_byte().await?;
            res |= (part as i32 & 0x7F) << (7 * i);
            if part & 0x80 == 0 {
                return Ok(res);
            }
        }
        error!("Varint is invalid");
        Err(std::io::ErrorKind::InvalidData.into())
    }

    // async fn read_varint(&mut self) -> Result<i32, NettyReadError> {
    //     let mut value = 0i32;
    //     let mut buf = [0u8; 1];
    //     let mut pos = 0u8;
    //     loop {
    //         self.read_exact(&mut buf).await?;
    //         println!("{}", buf[0]);
    //         value |= ((buf[0] & 0b01111111) << pos) as i32;
    //         if (buf[0] & 0b10000000) == 0 {
    //             break;
    //         };
    //         pos += 7;
    //         if pos >= 32 {
    //             return Err(std::io::ErrorKind::InvalidData.into());
    //         };
    //     }
    //     Ok(value)
    // }

    async fn read_varlong(&mut self) -> Result<i64, NettyReadError> {
        let mut value = 0i64;
        let mut buf = [0u8; 1];
        let mut position = 0u8;
        loop {
            self.read_exact(&mut buf).await?;
            value |= ((buf[0] & 0b01111111) << position) as i64;
            if (buf[0] & 0b10000000) == 0 {
                break;
            };
            position += 7;
            if position >= 64 {
                return Err(std::io::ErrorKind::InvalidData.into());
            };
        }
        Ok(value)
    }

    async fn read_vec3(&mut self) -> Result<(i32, i16, i32), NettyReadError> {
        let mut buf = [0u8; 8];
        self.read_exact(&mut buf).await?;
        let packed = u64::from_be_bytes(buf);
        let x: i32 = ((packed & 0xffffffc000000000) >> 38) as _;
        let y: i16 = (packed & 0xfff) as _;
        let z: i32 = ((packed & 0x3ffffff000) >> 12) as _;
        Ok((
            match x {
                i32::MIN..=0x1ffffff => x,
                0x2000000.. => x - 0x2000000,
            },
            match y {
                i16::MIN..=0x7ff => y,
                0x800.. => y - 0x800,
            },
            match z {
                i32::MIN..=0x1ffffff => z,
                0x2000000.. => z - 0x2000000,
            },
        ))
    }

    async fn read_uuid(&mut self) -> Result<u128, NettyReadError> {
        let mut buf = [0u8; 16];
        self.read_exact(&mut buf).await?;
        Ok(u128::from_be_bytes(buf))
    }

    async fn read_packet(&mut self) -> Result<Vec<u8>, NettyReadError> {
        let len = self.read_varint().await?;
        let mut buf = vec![0u8; len as usize];
        if len == 254 {
            let mut temp = [0u8];
            self.read_exact(&mut temp).await?;
            if temp[0] == 0xFA {
                // FE 01 FA: Legacy ServerListPing
                return Err(NettyReadError::LegacyServerListPing);
            }
            buf[0] = temp[0];
            self.read_exact(&mut buf[1..]).await?;
        } else {
            self.read_exact(&mut buf).await?;
        }
        Ok(buf)
    }

    // fn read_packet_compressed(&mut self) -> Result<Vec<u8>, NettyReadError> {
    //     let len = self.read_varint()?;
    //     let len_decompressed = self.read_varint()?;
    //     let mut buf = vec![0u8; len as usize];
    //     self.read_exact(&mut buf)?;
    //     if len_decompressed == 0 {
    //         return Ok(buf);
    //     }
    //     let mut buf_decompressed = vec![0u8; len_decompressed as usize];
    //     if flate2::Decompress::new(true)
    //         .decompress(&buf, &mut buf_decompressed, flate2::FlushDecompress::Finish)
    //         .is_err()
    //     {
    //         return Err(std::io::ErrorKind::InvalidData.into());
    //     };
    //     Ok(buf_decompressed)
    // }
}

impl<T: AsyncReadExt + Unpin> ReadExtNetty for T {}

#[async_trait]
pub trait WriteExtNetty: AsyncWriteExt + Unpin {
    async fn write_varint(&mut self, mut val: i32) -> std::io::Result<()> {
        for _ in 0..5 {
            if val & !0x7F == 0 {
                self.write_all(&[val as u8]).await?;
                return Ok(());
            }
            self.write_all(&[(val & 0x7F | 0x80) as u8]).await?;
            val >>= 7;
        }
        Err(std::io::ErrorKind::InvalidData.into())
    }

    async fn write_string(&mut self, s: &str) -> std::io::Result<()> {
        self.write_varint(s.len() as i32).await?;
        self.write_all(s.as_bytes()).await?;
        Ok(())
    }
}

impl<T: AsyncWriteExt + Unpin> WriteExtNetty for T {}
